"""add project_resource_policy.max_network_count column

Revision ID: 0bb88d5a46bf
Revises: 3215bc85dc45
Create Date: 2024-09-23 06:48:16.336099

"""

import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = "0bb88d5a46bf"
down_revision = "3215bc85dc45"
branch_labels = None
depends_on = None


def upgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    op.add_column(
        "project_resource_policies",
        sa.Column(
            "max_network_count", sa.Integer(), nullable=False, default="3", server_default="3"
        ),
    )
    # ### end Alembic commands ###


def downgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_column("project_resource_policies", "max_network_count")
    # ### end Alembic commands ###
