"""remove unique constraint of endpoint.url

Revision ID: 8b2ec7e3d22a
Revises: c5ed277b7f7b
Create Date: 2024-01-11 11:02:39.253292

"""

from alembic import op

# revision identifiers, used by Alembic.
revision = "8b2ec7e3d22a"
down_revision = "c5ed277b7f7b"
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_constraint("uq_endpoints_url", "endpoints", type_="unique")
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.create_unique_constraint("uq_endpoints_url", "endpoints", ["url"])
    # ### end Alembic commands ###
