"""rename desired_session_count to replicas

Revision ID: 428215fddea3
Revises: 6e44ea67d26e
Create Date: 2024-12-16 17:53:45.399528

"""

from alembic import op

# revision identifiers, used by Alembic.
revision = "428215fddea3"
down_revision = "6e44ea67d26e"
branch_labels = None
depends_on = None


def upgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    op.alter_column("endpoints", "desired_session_count", new_column_name="replicas")
    # ### end Alembic commands ###


def downgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    op.alter_column("endpoints", "replicas", new_column_name="desired_session_count")
    # ### end Alembic commands ###
